// Mini-Kompressor/Decompressor nach dem LZFG-Verfahren
// Aufruf: co <Eingabedatei> <Ausgabedatei>
// * extended header by Frank Gtze 2001 and change to COP.C (**)
// * fix in org-length & org-crc handling (size,alignment,endianess)
//   signature changed to declare this

#include <stdio.h>
#include <string.h>
#include <io.h>
#include <errno.h>

// #define ANALYZ
// #define VERBOSE
// #define SHIFT
// #define DEBUG
// #define X86_OPTIMIZED

/*
************************************************************
*                    ENDIAN CONVENTION
************************************************************
*/
 #define  LITTLE_ENDIAN   3412
 #define  BIG_ENDIAN      1234

 #define  BYTE_ORDER  LITTLE_ENDIAN   /* endianess of CPU */

 #define swap_l(n)          (((n & 0xff000000U) >> 24) | ((n & 0x00ff0000U) >> 8) | ((unsigned long)(n & 0x0000ff00U) << 8) | ((unsigned long)(n & 0x000000ffU) << 24))
 #define swap_s(n)          (((n & 0xff00) >> 8) | ((unsigned short)(n & 0x00ff) << 8))
 
//------------ for little-endian protocols -----------------
 #if BYTE_ORDER == LITTLE_ENDIAN
  #define L_htonl(n)       n
  #define L_htons(n)       n
 #else
  #define L_htonl(n)       swap_l(n)
  #define L_htons(n)       swap_s(n)
 #endif
 #define  L_ntohl          L_htonl
 #define  L_ntohs          L_htons
//-------------- for big-endian protocols ------------------
 #if BYTE_ORDER == BIG_ENDIAN
  #define B_htonl(n)       n
  #define B_htons(n)       n
 #else
  #define B_htonl(n)       swap_l(n)
  #define B_htons(n)       swap_s(n)
 #endif
 #define  B_ntohl          B_htonl
 #define  B_ntohs          B_htons
 
/*
************************************************************
*                   COP-LZFG CONVENTIONS
************************************************************
*/
#define KB  1024
#define FG  (4*KB-CL)           /* Fenstergre        */
#define LL  16                  /* max. Literallnge   */
#define CL  16                  /* max. Copy-Lnge     */
#define NIL 0xFFFF8000          /* 'nichts'            */

#define TICKER      (*(unsigned short volatile _far*)0x0040006C)
#define AddLiteral  { ShiftWindow (1); ++l; }
#define ModIndex(i) ((i)&(sizeof(Window)-1))
#define AccIndex(i) (ModIndex(b+(i)))

#ifdef ANALYZ
 #define Append(c) { PutPrintable(Window[t++]=c); t &= sizeof(Window)-1; }
#else
 #define Append(c) { fputc(Window[t++]=c,ofile); t &= sizeof(Window)-1; }
#endif

static  unsigned char  Window[FG+CL];            // Fenster in die Vergangenheit
static  unsigned short Link[sizeof(Window)];     // Verkettungsspeicher
static  long Head[256];                          // Listenkpfe
static  long Start;                              // Offset fr 'Window'

static  int m;                                   // Match-Lnge 0...CL
static  int l;                                   // Literallnge 0...LL
static  int p;                                   // letzte Matchposition
static  int t;                                   // Lnge des aktuellen Strings
static  int b;                                   // Buffer Base Position

static  int ixf, ixq;                            // window indeces at FG, q


/*
   'Window' wird wie folgt benutzt:
    Window[0..FG-l-1]   bereits gesendet (am Anfang initialisiert)
    Window[FG-l..FG-1]  ist das nchste zu sendende Literal
    Window[FG..FG+m-1]  Kopie (Match) von Window[p..p+m-1],
                        die eventuell noch verlngert werden kann
    Window[FG..FG+t-1]  aktuell zu untersuchende Zeichenkette
*/

static  char Signat[] = "LZFG\x17\x57";     // fr komprimierte Daten plus extended header (**)
static  long ibytes = 0, obytes = 0;        // Statistik
static  FILE *ifile, *ofile;                // Eingabe, Ausgabe

/************************************************************************/
#pragma check_stack(off)

static void Usage(void)
{
    static  char usage[] =  "\nusage:\tco file compressed_file"
                            "\n\tco file [expanded_file]"
                            "\n\n\ttype of 'file' is detected automatically"
                            "\n\tconsole is default 'expanded_file'\n";

    fprintf(stderr,usage);
    exit(1);
}

static void DecERR(void)                    // new FG (**)
{
    static char decERR[] = "\nlenght-error by uncompressing file !\n";

    fprintf(stderr,decERR);
    exit(1);
}

static void crcERR(void)                    // new FG (**)
{
    static char crcERR[] = "\nCRC-error by uncompressing file !\n";

    fprintf(stderr,crcERR);
    exit(1);
}

unsigned short cop_CRC(FILE *cfile, unsigned long clenght)// new FG (**)
{                                           // use lenght for correct CRC
    unsigned char  chr;
    unsigned short crc, i;
    unsigned long  bytes;

    crc   = 0;
    bytes = 0;
    while(bytes++ < clenght) {
        chr  = getc(cfile);
        crc ^= (unsigned short)(chr << 8);
        for(i=0; i<8; i++) {
            if(crc & 0x8000)
                crc = (crc << 1) ^ 0x1021;
            else
                crc <<= 1;
        }
    }
    return(crc);
}

#ifdef VERBOSE
static void PutPrintable(int chr)
{
    static char escape[] = { 0,0,0,0,0,0,0,'a','b','t','n',0,'f','r' };

    if(chr < 0x20) {
        fputchar('\\');
        if(chr < sizeof(escape) && escape[chr])
            fputchar(escape[chr]);
        else
            printf("x%.2x",chr);
    } 
    else if(chr == '\\')
        printf("\\\\");
    else
        fputchar(chr);
}
#endif	// VERBOSE

#ifdef ANALYZ
static void PutPrintable(int chr)
{
    static char escape[] = { 0,0,0,0,0,0,0,'a','b','t','n',0,'f','r' };

    if(chr < 0x20) {
        fputchar('\\');
        if(chr < sizeof(escape) && escape[chr])
            fputchar(escape[chr]);
        else
            printf("x%.2x",chr);
    } 
    else if(chr == '\\') 
        printf("\\\\");
    else
        fputchar(chr);
}
#endif    // ANALYZ

static void InitWindow(void)                // Fenster initialisieren
{
    int i;

    for(i=0; i<sizeof(Window); ++i) Window[i] = 0, Link[i] = NIL;
    for(i=0; i<sizeof(Head)/sizeof(long); ++i) Head[i] = NIL;
    Start = m = l = t = 0;
    b = 0;
    p = NIL;                                // Anfangswerte
}

static void ShiftWindow(int i)              // schiebe i Zeichen nach links
{
    register int link;

    m  = 0;                                 // Match ist leerer String
    t -= i;                                 // Zeichenpuffer top Position
    b += i;                                 // Basisindex fr Ringpuffer
    Start += i;                             // Window offset korrigieren
#ifdef SHIFT
    {
        int ix, j, k;

        printf("shift %d: ",i);
        for(j=sizeof(Window)-48; j<sizeof(Window); ++j)
            fputchar(Window[AccIndex(j)]);
        fputchar('\n');
        printf("p = %d\n",p);
        for(j=0; j<256; ++j)
            if(Head[j] != NIL) {
                printf("Head[%c] = %ld:",j,Head[j]);
                if((k = Head[j] - Start) >= 0) {           // latest occurence was here
                    do {
                        ix = AccIndex(k);
                        printf(" %2d(%c)",k,Window[ix]);
                    } while((k = k - Link[ix]) >= 0);
                }
                fputchar('\n');
            }
        fputchar('\n');
    }
#endif	// SHIFT
    link = Link[AccIndex(FG)];
    p    = (t != 0 && link != NIL)? FG - link : NIL;
}

static void SendLiteral(void)               // literal Window[FG-l..FG-1] senden
{
    int ix;

    if(l) {
#ifdef DEBUG
        putchar('.');
#else
 #ifdef VERBOSE
        printf("%.2d bytes literal >|",l);
 #endif
#endif
        obytes += 1 + l;                    // Lnge der Ausgabedatei
        putc(l-1,ofile);                    // Lnge ausgeben '0000xxxx'
        while(l) {
#ifdef VERBOSE
            PutPrintable(Window[AccIndex(FG-l)]);
#endif
            putc(Window[AccIndex(FG-l--)],ofile); // Zeichen fr Zeichen...
        }
#ifdef VERBOSE
        fputchar('\n');
#endif
    }
}

static void SendCopy(void)                  // Window[FG..FG+m-1] als Rckbezug
{
    register int ix, cnt;

    SendLiteral();                          // vorher noch literalen string
    if(cnt = m) {
        ix = FG - 1 - p;
#ifdef DEBUG
        putchar('+');
#else
 #ifdef VERBOSE
        printf("%.2d bytes [-%.4X] >|",m,ix);
 #endif
#endif
        putc(((m-1)<<4) | (ix>>8),ofile);
        putc(ix & 0xff,ofile);
        obytes += 2;                        // 4 bit Lnge, 12 bit Offset
#ifdef VERBOSE
        ix = AccIndex(FG-1-ix);
        while(cnt--) {
            PutPrintable(Window[ix]);
            ix = ModIndex(++ix);
        }
        fputchar('\n');
#endif
        ShiftWindow(m);
    }
}

#pragma	check_stack (on)
/************************************************************************/


#ifndef X86_OPTIMIZED                       // see below for equivalent code (x86-asm).....
static int MatchExtends(void)
// Kommt die Zeichenkette Window[FG..FG+m] im Fenster vor?
// Vorbedingung: Window[FG..FG+m-1] kommt an Position p vor
{
    register int ixf, ixq;
    int n, q, ixb;

    if(p < 0) return(0);                    // kein Match mglich

    ixq = AccIndex (p+m);
    ixf = AccIndex (FG+m);
    if(Window[ixq] == Window[ixf]) return(1);
                                            // Match an alter Position p verlngert
    ixb = AccIndex(q = p);                  // Match an neuer Position q ?
    while((q = q-Link[ixb]) >= 0) {
        ixf = AccIndex(FG);
        ixq = AccIndex(q);
        ixb = ixq;
        n = 0;
        while(Window[ixf] == Window[ixq])   // solange gleich
            if(n == m) { p = q; return(1); }
            else {
                ++n;
                ixf = ModIndex (++ixf);
                ixq = ModIndex (++ixq);
            }
    }
    return(0);
}
#endif	// X86_OPTIMIZED

void Compress(void)
{
    int  chr, index, ofset;
    long save;

    InitWindow();
    for(;;) {
        if(l == LL) SendLiteral();          // Lngenlimit LL
        if(m == CL) SendCopy();             // Lngenlimit CL
        if(m == t) {
            if((chr = getc (ifile)) == EOF) break;
            ofset = AccIndex(index = FG+t);
            Window[ofset] = chr;            // Fenster nachfllen
            save = Head[chr];
            Head[chr] = Start + index;
            if(save != NIL) {
                Link[ofset] = save = Head[chr] - save;
                if(save > FG) Link[ofset] = save = NIL;
                else if(t == 0) p = index - Link[ofset];
            }
            ++t;                            // aktuelle Lnge
            ++ibytes;                       // Bytes mitzhlen
        }
#ifdef X86_OPTIMIZED
_asm {   cmp    p,0                 ; try to extend current match..
         jge    DoCheck
         jmp    NoMatch             ; no match possible

DoCheck: mov    si,b                ; look if old position
         add    si,m                ; still matches
         mov    di,si               ; b + m
         sub    si,16
         and    si,0FFFH            ; ixf = (b+m+FG) mod (4*K)
         add    di,p
         and    di,0FFFH            ; ixp = (b+m+p)  mod (4*K)
         mov    al,Window[si]
         cmp    al,Window[di]
         je     MatchExtend

         mov    dx,p                ; try to find a new match
         mov    bx,dx               ; back in history
         add    bx,b
         and    bh,0FH

TryAgain: 
         shl    bx,1
         mov    ax,Link[bx]
         sub    dx,ax               ; loop test
         js     NoMatch             ; exit if < 0

         mov    si,b
         mov    di,si               ; preset with window base
         sub    si,16               ; equivalent to +FG
         and    si,0FFFH            ; modulo window size
         add    di,dx
         and    di,0FFFH            ; q mod (window size)
         mov    bx,di               ; save for later

         mov    cx,m                ; byte counter
         inc    cx                  ; must process m+1 bytes(!)
InnerLoop:
         mov    al,Window[si]
         cmp    al,Window[di]       ; compare bytes and exit
         jne    TryAgain            ; loop when different
         inc    si
         and    si,0FFFH
         inc    di
         and    di,0FFFH            ; increment pointers mod (size)
         loop   InnerLoop           ; to compare next item in loop

         mov    p,dx                ; save head of reference string
MatchExtend:
         inc    m                   ; count size for this match
} continue;                                 // process next character
NoMatch:
#else
        if(MatchExtends()==1)  m++;         // process next character
        else                                // !!! no brackets here !!!
#endif
        if(m > 2 || (m == 2 && l == 0))
            SendCopy();                     // Rckbezug senden
        else
            AddLiteral;                     // Zeichen Window[FG] zum Literal
    }
    if(m == 1) AddLiteral;                  // bugfix for orginal
    if(m < 2) l += m, m = 0;
    SendCopy();                             // EOF: der letzte Rest
}

void Expand(void)
{
    int len, off, c;

    InitWindow();
    while((c = fgetc (ifile)) != EOF) {
        ++ibytes;
        if(c & 0xf0) {                      // Rckbezug
            len = 1 + (c>>4);
            off = t - (((c & 0x0f)<<8) + fgetc(ifile)) - 1;
            ++ibytes;
#ifdef ANALYZ
            printf("\n@%.6lX: %.2d bytes [-%.4X] >|",obytes,len,t-off-1);
#endif
            obytes += len;
            while(len--) {
                off &= sizeof(Window)-1;    // modulo Puffer
                Append(Window[off++]);      // Zeichen aus Puffer
            }
        } else {                            // Literal
            len = 1 + c;                    // 0..15 => 1..16
#ifdef ANALYZ
            printf("\n@%.6lX: %.2d bytes literal >|",obytes,len);
#endif
            obytes += len;
            while(len--) { Append (fgetc(ifile)); ++ibytes; }
        }
    }
}

/****************************** MAIN ************************************/
main(int argc,char *argv[])
{
    unsigned short time;
    long ratio;

    long fsize, fsizet;
    int  filenr;
    unsigned short crci, crco, crct;

    char Challenge[sizeof(Signat)];

    if(argc < 2) Usage();                   // Benutzerhilfe?
    ifile = fopen(argv[1],"rb");
    fgets(Challenge,sizeof(Challenge),ifile);
    if(memcmp(Challenge,Signat,sizeof(Signat)-1)) {
        if(argc < 3) Usage();               // outfile is mandatory
        ofile = fopen(argv[2],"wb");
        rewind(ifile);                      // compression mode
        filenr = fileno(ifile);             // (**) get file-number for next
        fsize  = filelength(filenr);        // (**) get lenght of org-file
        crci   = cop_CRC(ifile,fsize);      // (**) make CRC-16 over org-file
        printf("\norg-size: %lu  org-crc: %i\n",fsize,crci);  // (**) check-display
        fputs(Signat,ofile);                // write signatur
        fsizet = B_htonl(fsize);            // create big-endian infos
        crct   = B_htons(crci);
        fwrite(&fsizet,1,sizeof(long),ofile);  // (**) write org-filelenght as big-endian
        fwrite(&crct,1,sizeof(short),ofile);   // (**) write org-CRC16 as big-endian
        rewind(ifile);                      // (**) reset to begin
        ibytes = 0;
        obytes = sizeof(Signat)+sizeof(fsize)+sizeof(crci);
                                            // (**) change over new header-ext.
        printf("\ncompressing .....\n");
        time = TICKER;
        Compress();
        time = 10000L*(TICKER - time)/182;
        ratio = 100*(ibytes-obytes)/ibytes;
    } else {                                // expansion mode
        ofile = argv[2]? fopen(argv[2],"wb+") : stdout;
        obytes = ibytes = 0;                // (**) change for reading  (CRC)
#ifdef ANALYZ
        printf("\nanalyzing .......\n");
#else
        printf("\nexpanding .......\n");
#endif
        time = TICKER;
        fread(&fsizet,1,sizeof(long),ifile);  // (**) read org-filelenght from header as big-endian
        fread(&crct,1,sizeof(short),ifile);   // (**) read org-CRC16 from header as big-endian
        fsize = B_ntohl(fsizet);            // correct endianess from big-endian
        crci  = B_ntohs(crct);
        ibytes=sizeof(Signat)+sizeof(fsize)+sizeof(crci);
                                            // (**) change over new header-ext.
        Expand();
        rewind(ofile);                      // (**) reset to begin
        crco=cop_CRC(ofile,obytes);         // (**) make CRC over decompressed file
        printf("org-size: %lu  org-crc: %i\n",fsize,crci);   // (**) check-display
        printf("out-size: %lu  out-crc: %i\n",obytes,crco);  // (**)    -"- 
        if(fsize!=obytes) DecERR();         // (**) control correct lenght
        if(crci!=crco) crcERR();            // (**) control correct CRC
        time = 10000L*(TICKER - time)/182;
        ratio = 100*(obytes-ibytes)/obytes;
    }
    printf("in %ld out %ld (%ld%%)\n",ibytes,obytes,ratio);
    printf("elapsed time %u msec\n",time);
    if(time) printf("working at %lu bytes/sec in, %lu bytes/sec out\n", 
                     1000*ibytes/time,1000*obytes/time);
    fclose(ifile);
    fclose(ofile);
    return(0);
}
